"""
"""


#[
from __future__ import annotations
#]


class Variant:
    """
    """
    #[

    __slots__ = (
        "solution",
    )

    def __init__(
        self,
        order: int,
        constant: bool,
    ) -> None:
        """
        """
        self.solution = None

    #]

